from sack import __version__


SACK_ASCII = r"""
█▀ ▄▀█ █▀▀ █▄▀
▄█ █▀█ █▄▄ █ █
"""

SACK_ABOUT = f"""
version: {__version__}

sack is an open-source chat application for the terminal,
built with the Textual framework and TCP sockets.

sack is distributed under the MIT license."""

WELCOME_HELP = """\
The welcome screen is the first screen you see when you run sack.

From there, you can start or join a server,
view the about page, and open the help popup."""

SERVER_HELP = """\
On the 'Create server' form, you will be asked to enter a server port.

If the chosen port is available, a new server will be started
on your machine and visible to all devices connected to your local network.
You will then be asked to choose a nickname.

If you exit sack, your server will stop running
and all connected users will be disconnected."""

JOIN_HELP = """\
On the 'Join server' form, you will be asked to enter the server host and port.
If the server is reachable, you will be connected and then asked to choose a nickname.

Make sure you enter the correct address and port,
otherwise the connection will not be established."""

CHAT_HELP = """\
The chat screen consists of a message area and a text input field.

When the input field is focused, you can type and send messages
to other users.

When the message area is focused, you can use
the available keybindings to scroll through messages."""



WELCOME_KB = [
    ("j", "down"),
    ("k", "up"),
    ("enter", "confirm"),
    ("ctrl+c", "exit"),
    ("f1", "help"),
]

FORMS_KB = [
    ("ctrl+j", "down"),
    ("ctrl+k", "up"),
    ("enter", "confirm"),
    ("ctrl+c", "exit"),
    ("esc", "back"),
    ("f1", "help"),
]

ABOUT_KB = [
    ("ctrl+c", "exit"),
    ("esc", "back"),
    ("f1", "help"),
]

HELP_KB = [
    ("l", "next tab"),
    ("h", "previous tab"),
    ("j", "down"),
    ("k", "up"),
    ("esc", "back"),
]

CHAT_KB = [
    ("ctrl+j", "focus up"),
    ("ctrl+k", "focus down"),
    ("enter", "send message"),
    ("shift+enter", "new line"),
    ("ctrl+c", "exit"),
    ("esc", "menu"),
    ("f1", "help"),
]



import signal
import logging
import threading

from typing import Protocol
from argparse import ArgumentParser
from collections.abc import Sequence

from sack.models import SackClient, SackServer


def main() -> None:
    logging.basicConfig(level=logging.DEBUG, format="%(asctime)s | %(levelname)s | %(name)s: %(message)s")
    args = get_args()
    args.func(args)


def get_args(args: Sequence[str] | None = None):
    parser = ArgumentParser()
    parser.set_defaults(func=main)
    subparsers = parser.add_subparsers()

    server_subparser = subparsers.add_parser("server")
    server_subparser.add_argument("-p", "--port", type=int, required=False, default=8080)
    server_subparser.add_argument("--host", required=False, default="localhost")
    server_subparser.set_defaults(func=server_controller)

    client_subparser = subparsers.add_parser("client")
    client_subparser.add_argument("-p", "--port", type=int, required=False, default=8080)
    client_subparser.add_argument("--host", required=False, default="localhost")
    client_subparser.add_argument("--username", required=True)
    client_subparser.set_defaults(func=client_controller)

    arguments = parser.parse_args(args)
    return arguments


class ServerControllerArgs(Protocol):
    host: str
    port: int


def server_controller(args: ServerControllerArgs) -> None:
    def sigint_handler(*_):
        s.stop()

    signal.signal(signal.SIGINT, sigint_handler)

    with SackServer(args.host, args.port) as s:
        s.serve()


class ClientControllerArgs(Protocol):
    host: str
    port: int
    username: str


def client_controller(args: ClientControllerArgs) -> None:
    client = SackClient(host=args.host, port=args.port, username=args.username)
    client.connect()
    client.join_request()

    def listen():
        while True:
            msg = client.receive_message()
            if msg is None:
                continue
            if msg.username == args.username:
                continue
            if msg.type == "CONNECT":
                print(f"{msg.username} joined")
            if msg.type == "DISCONNECT":
                print(f"{msg.username} disconnected")
            if msg.type == "TEXT":
                print(f"{msg.username}: {msg.text}")

    listener = threading.Thread(target=listen, daemon=True)
    listener.start()

    while True:
        text = input()
        if text == "q":
            exit()
        client.send_text(text)



from multiprocessing import Process

from textual.app import App, ComposeResult
from textual.worker import Worker
from textual.binding import Binding
from textual.widgets import Label, Button
from textual.containers import Container

from sack.util import make_keybinding_text
from sack.models import AsyncSackClient
from sack.screens import (
    HelpScreen,
    AboutScreen,
    ThemeChangeScreen,
    ClientPromptScreen,
    ServerPromptScreen,
)
from sack.components import Option, Options, SackHeader
from sack.keybindings import WELCOME_KB


class SackApp(App):
    CSS_PATH = "styles.css"
    SCREENS = {
        "1": ServerPromptScreen,
        "2": ClientPromptScreen,
        "3": ThemeChangeScreen,
        "4": HelpScreen,
        "5": AboutScreen,
    }
    BINDINGS = [
        Binding("ctrl+c", "exit"),
        Binding("ctrl+t", "push_screen('3')"),
        Binding("f1", "push_screen('4')"),
    ]
    ENABLE_COMMAND_PALETTE = False

    def __init__(self):
        super().__init__()
        self.HEADER_BREAKPOINT = 20
        self.server_process: Process | None = None
        self.client: AsyncSackClient | None = None
        self.message_worker: Worker | None = None

    def compose(self) -> ComposeResult:
        yield from self.get_header()
        with Options(id="options"):
            yield Option("Create server", "server")
            yield Option("Join server", "join")
            yield Option("Help", "help")
            yield Option("About sack", "about")
            yield Option("Exit", "exit")
        with Container(id="footer-container"):
            yield Label(make_keybinding_text(WELCOME_KB), classes="container")

    def action_exit(self):
        self.exit()

    def on_resize(self, _):
        height = self.size.height
        header = self.query_one(SackHeader)
        header.display = height >= self.HEADER_BREAKPOINT

    def on_button_pressed(self, e: Button.Pressed):
        match e.button.id:
            case "server":
                self.push_screen("1")
            case "join":
                self.push_screen("2")
            case "help":
                self.push_screen("4")
            case "about":
                self.push_screen("5")
            case "exit":
                self.exit()

    def get_header(self):
        header = SackHeader()
        header.display = self.size.height >= self.HEADER_BREAKPOINT
        yield header

    async def cleanup(self):
        if self.message_worker:
            self.message_worker.cancel()
            self.message_worker = None
        if self.client:
            await self.client.disconnect()
            self.client = None
        if self.server_process:
            self.server_process.kill()
            self.server_process = None

    def back_to_first_screen(self):
        for _ in range(len(self.screen_stack) - 1):
            self.pop_screen()


def main(*_):
    app = SackApp()
    app.run()


if __name__ == "__main__":
    main()



from .client import (
    SackClient,
    AsyncSackClient,
    SackClientError,
    SackClientServerError,
    SackClientUsernameError,
)
from .server import SackServer
from .protocol import SackMessage


__all__ = [
    "SackClient",
    "AsyncSackClient",
    "SackClientError",
    "SackClientServerError",
    "SackClientUsernameError",
    "SackServer",
    "SackMessage",
]



import socket
import asyncio

from sack.models.protocol import SackMessage, receive_message


class SackClientError(Exception):
    pass


class SackClientServerError(SackClientError):
    pass


class SackClientUsernameError(SackClientError):
    pass


# todo username setter
class SackClient:
    def __init__(self, *, host: str, port: int, username: str | None = None) -> None:
        self.host = host
        self.port = port
        self.username = username
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    def connect(self) -> None:
        try:
            self._socket.connect((self.host, self.port))
        except Exception as e:
            raise SackClientServerError from e

    def join_request(self) -> None:
        assert self.username
        msg = SackMessage("CONNECT", self.username)
        self._socket.sendall(msg.to_bytes())
        ok_no = self._socket.recv(2).decode()
        if ok_no == "NO":
            self.disconnect()
            raise SackClientUsernameError

    def disconnect(self) -> None:
        self._socket.shutdown(socket.SHUT_RDWR)
        self._socket.close()

    def send_text(self, text: str) -> None:
        assert self.username
        msg = SackMessage("TEXT", self.username, text)
        self._socket.sendall(msg.to_bytes())

    def receive_message(self) -> SackMessage | None:
        def on_empty():
            raise SackClientServerError

        return receive_message(self._socket, on_empty)


class AsyncSackClient:
    def __init__(self, *, host: str, port: int, username: str | None = None) -> None:
        self.host = host
        self.port = port
        self.username = username

    async def connect(self, *, timeout: float | None = None) -> None:
        try:
            async with asyncio.timeout(timeout):
                reader, writer = await asyncio.open_connection(self.host, self.port)
        except Exception as e:
            raise SackClientServerError from e
        self._reader, self._writer = reader, writer

    async def join_request(self) -> None:
        assert self.username
        msg = SackMessage("CONNECT", self.username)
        self._writer.write(msg.to_bytes())
        await self._writer.drain()
        ok_no = await self._reader.read(2)
        ok_no = ok_no.decode()
        if ok_no == "NO":
            raise SackClientUsernameError

    async def disconnect(self) -> None:
        self._writer.close()
        await self._writer.wait_closed()

    async def send_text(self, text: str) -> None:
        assert self.username
        msg = SackMessage("TEXT", self.username, text)
        self._writer.write(msg.to_bytes())
        await self._writer.drain()

    async def receive_message(self) -> SackMessage | None:
        message_len = await self._reader.read(1)
        if not message_len:
            raise SackClientServerError
        message_len = int.from_bytes(message_len, "big")
        raw_message = await self._reader.read(message_len)
        raw_message = raw_message.decode()
        message_parts = raw_message.split("\n")
        if not len(message_parts) == 2:
            return None
        type = message_parts[0]
        username = message_parts[1]
        if type not in ("CONNECT", "TEXT", "DISCONNECT"):
            return None
        if type != "TEXT":
            return SackMessage(type, username)
        sep = await self._reader.read(1)
        if sep != b"\n":
            return None
        text_len = await self._reader.read(2)
        text_len = int.from_bytes(text_len, "big")
        text = await self._reader.read(text_len)
        text = text.decode()
        return SackMessage(type, username, text)



import random

from collections.abc import Collection

from textual.color import Color


class ColorsManager:
    def __init__(self) -> None:
        self._registry: dict[str, Color] = {}
        self.color_stack = [
            Color.parse("#1E90FF"),  # blue
            Color.parse("#DC143C"),  # red
            Color.parse("#FFFF00"),  # yellow
            Color.parse("#7FFF00"),  # green
            Color.parse("#8A2BE2"),  # violet
            Color.parse("#FF1493"),  # pink
        ]
        random.shuffle(self.color_stack)

    def get(self, username: str) -> Color:
        if username in self._registry:
            return self._registry[username]
        if self.color_stack:
            color = self.color_stack.pop()
        else:
            color = Color(
                random.randint(0, 255),
                random.randint(0, 255),
                random.randint(0, 255),
            )
        self._registry[username] = color
        return color


def make_keybinding_text(keybindings: Collection[tuple[str, str]]):
    return "  ".join(f"[$secondary]{key}[/] {desc}" for key, desc in keybindings)



from typing import Any, Literal
from datetime import datetime

from textual import events
from textual.app import ComposeResult
from textual.color import Color
from textual.widget import Widget
from textual.binding import Binding
from textual.widgets import Input, Label, Button, Static, TextArea
from textual.containers import Center, Container, VerticalScroll, HorizontalGroup

from sack.assets import SACK_ASCII


class SackHeader(HorizontalGroup):
    def compose(self) -> ComposeResult:
        yield Static(classes="filler")
        yield Static(SACK_ASCII, id="header", classes="container")
        yield Static(classes="filler")


class TextInput(TextArea):
    def _on_key(self, event: events.Key) -> None:
        new_line = "shift+enter"
        if event.key != new_line:
            return

        event.stop()
        event.prevent_default()
        start, end = self.selection
        self._replace_via_keyboard("\n", start, end)


class ChatMessage(Widget):
    def __init__(
        self,
        orientation: Literal["left", "right"],
        msg: str,
        author: str,
        color: Color | None = None,
    ):
        super().__init__()
        self.orientation = orientation
        self.msg = msg
        self.author = author
        self.add_class("msg-container")
        self.add_class(self.orientation)
        self.color = color

    def compose(self) -> ComposeResult:
        container = Container(classes="msg")
        if self.color:
            setattr(container.styles, f"outline_{self.orientation}", ("solid", self.color))
        with container:
            yield Label(self.msg, classes="msg-text")
            yield Label(f"{self.author} ({datetime.now().strftime('%H:%M')})", classes="msg-author")


class Options(Container):
    BINDINGS = [
        Binding("down, tab, j", "app.focus_next"),
        Binding("up, shift+tab, k", "app.focus_previous"),
    ]


class Option(HorizontalGroup):
    def __init__(self, label: str, option_key: str) -> None:
        super().__init__()
        self.label = label
        self.option_key = option_key
        self.add_class("container")

    def compose(self) -> ComposeResult:
        yield Label(">", classes="option-arrow")
        yield Label(self.label, classes="option-label")
        yield Button(id=self.option_key)


class VimVerticalScroll(VerticalScroll):
    BINDINGS = [
        Binding("k", "scroll_up", "Scroll Up", show=False),
        Binding("j", "scroll_down", "Scroll Down", show=False),
        Binding("g", "scroll_home", "Scroll Home", show=False),
        Binding("G", "scroll_end", "Scroll End", show=False),
        Binding("u", "page_up", "Page Up", show=False),
        Binding("d", "page_down", "Page Down", show=False),
    ]


class FormErrors(Center):
    def __init__(self):
        super().__init__()
        self._errors: dict[int, str] = {}

    def compose(self) -> ComposeResult:
        label = Label(classes="form-error")
        label.display = False
        yield label

    def reset(self) -> None:
        self._errors = {}
        self._update(display=False)

    def set_error(self, err_id: int, err_message: str):
        self._errors[err_id] = err_message
        self._update(display=True)

    def clear_error(self, err_id: int) -> None:
        if err_id not in self._errors:
            return
        del self._errors[err_id]
        self._update(display=bool(self._errors))

    def has_errors(self, *err_ids: int) -> bool:
        if not err_ids:
            return bool(self._errors)
        for err_id in err_ids:
            if err_id in self._errors:
                return True
        return False

    def _update(self, *, display: bool) -> None:
        label = self.query_one(".form-error", Label)
        label.update("\n".join(self._errors.values()))
        label.display = display


class FormField(HorizontalGroup):
    def __init__(self, id: str, label: str, **input_kwargs: Any) -> None:
        super().__init__()
        self.input_id = id
        self.label = label
        self.input_kwargs = input_kwargs

    def compose(self) -> ComposeResult:
        yield Label("> ", classes="option-arrow")
        yield Label(self.label, classes="form-label")
        yield Input(id=self.input_id, compact=True, **self.input_kwargs)  # type: ignore

    @property
    def value(self) -> str:
        return self.query_one(Input).value


class FormButton(HorizontalGroup):
    def __init__(self, label: str) -> None:
        super().__init__()
        self.label = label

    def compose(self) -> ComposeResult:
        yield Label("> ")
        yield Button(self.label, compact=True)


class HelpTab(Container):
    def __init__(self, label: str, tab_id: str) -> None:
        super().__init__()
        self.label = label
        self.tab_id = tab_id

    def compose(self) -> ComposeResult:
        yield Label(self.label)
        yield Button(id=self.tab_id)


class HelpKeybinding(HorizontalGroup):
    def __init__(self, key: str, desc: str) -> None:
        super().__init__()
        self.key = key
        self.desc = desc

    def compose(self) -> ComposeResult:
        yield Label(self.key, classes="key")
        yield Label(self.desc, classes="desc")
